/***{{{*******************************************************************
 * This file is part of libco - object library for C                     *
 * Copyright (c) 2002                                                    *
 *     Peter Rockai (yenar) <yenar@host.sk>                              *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2 of the License, or (at your option) any later version.      *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 59 Temple Place, Suite 330,                         *
 * Boston, MA 02111-1307  USA                                            *
 *******************************************************************}}}***/
/* {{{ file description */
/** @file co_object.h
 *  @brief Object definition and function prototypes. */
/* }}} */

#ifndef __CO_OBJECT_H__
#define __CO_OBJECT_H__

#include <stdarg.h>
#include "co_type.h"
#include "co_exception.h"
#include "co_string.h"
#include "co_debug.h"

#if DEBUGLVL > 0
#include <stdio.h>
#endif

/* {{{ co message and event macros */
/// Send message to object @a o (message selector is first arg).
#define CO_M(o, ...) (CO_DEBUG (8, "sending msg..."), \
		co_m (o, o, COM (__VA_ARGS__)))
#define CO_D(o, ...) (co_m (o, this, COM (__VA_ARGS__)))
#define CO_D_V(o, s, m, v) (co_m_v (o, s, m, v))
#define CO_A(t, o, ...) (CO_DEBUG (8, "getting attribute..."), \
		* (co_m (o, o, COM (__VA_ARGS__)) . _ ## t ## _p))
/// Send message to object @a o (message selector is first arg).
#define CO_M_V(o, a, va) (CO_DEBUG (8, "sending msg..."), \
		co_m_v (o, COM (a), va))

/// Raise event in object @a o.
#define CO_EV(o, ...) (co_ev (o, COE (__VA_ARGS__)))

/// Connect object @a a event (signal) to object @a b message (slot)
#define CO_CNT(a, e, b, m) (co_cnt ((a), (COE (e)), b, COM (m), false))
#define CO_CNT_AD(a, e, b, m) (co_cnt ((a), (COE (e)), b, COM (m), true))
/* }}} */
/* {{{ co miscelaneous macros */
#define CO_P(o, p, t) (*(((((co_o) (o)) -> prop_get) (o, COP (p))) . t))
#define CO_P_S(o, p, t, fb) \
	((o)?(CO_P (o, p, t)):(fb))
#define CO_IS(o, _t) ((((co_o) (o)) -> t) == COT (_t))
#define CO_PUSH(o) (cox_push ((o), cox_co_cleanup, 0))
/* }}} */
/* {{{ co creation macros */
#define CO_BLESS(o, t, m) (co_bless ((o), COT (t), (m)))
#define CO_BLESS_0(o, t) CO_BLESS (o, t, 0)
#define CO_BLESS_M(o, t) CO_BLESS (o, t, t ## _msg)
#define CO_BLESS_P(o, t) CO_BLESS (o, t, 0)
#define CO_BLESS_MP(o, t) CO_BLESS (o, t, t ## _msg, t ## _p)
/// Create and bless object.
#define CO_NEW(t, ...) (__co_o_bless (t ## _new (__VA_ARGS__)))
#define CO_NEW_P(t, ...) (__co_o_bless_p (t ## _new (__VA_ARGS__)))
#define CO_F_NEW(t, ...) (__co_o_bless ((co_f_get (t)) (__VA_ARGS__)))
#define CO_F_NEW_P(t, ...) (__co_o_bless_p ((co_f_get (t)) (__VA_ARGS__)))
#define CO_FS_NEW(t, ...) (__co_o_bless ((co_fs_get (t)) (__VA_ARGS__)))
/* }}} */
/* {{{ co id construction macros */
/// Construct real co action identifier.
#define COM(...) __co_id_message_ ## __VA_ARGS__
/// Construct real co type identifier.
#define COT(...) __co_id_type_ ## __VA_ARGS__
/// Construct real co event identifier.
#define COE(...) __co_id_event_ ## __VA_ARGS__
/// Construct real co property identifier.
#define COP(...) __co_id_property_ ## __VA_ARGS__
/* }}} */
/* {{{ co id initialization macros */
/// Declaration of event ID. Use in global library header file.
#define COE_DECL(x) extern unsigned __co_event_ ## x
/// Definition of event ID. Use in global part of .c file.
#define COE_DEF(x) unsigned __co_event_ ## x
/// Initialization of event ID. Use in static library init function.
#define COE_NEW(x) ((void)((((COE (x)) = co_event_new (#x)), 1) && \
	((CO_DEBUG (1, "registered co event: " #x " = %d", COE (x))), 1)))

/// Declaration of action ID. Use in global library header file.
#define COM_DECL(x) extern unsigned __co_action_ ## x
/// Definition of action ID. Use in global part of .c file.
#define COM_DEF(x) unsigned __co_action_ ## x
/// Initialization of action ID. Use in static library init function.
#define COM_NEW(x) ((void)((((COM (x)) = co_action_new (#x)), 1) && \
	((CO_DEBUG (1, "registered co action: " #x " = %d", COM (x))), 1)))

/// Declaration of type ID. Use in global library header file.
#define COT_DECL(x) extern unsigned __co_type_ ## x; CO_TYPEDEF (x)
#define COT_DECL_V(x) extern unsigned __co_type_ ## x;
/// Definition of type ID. Use in global part of .c file.
#define COT_DEF(x) unsigned __co_type_ ## x
/// Initialization of type ID. Use in static library init function.
#define COT_NEW(x) ((void)((((COT (x)) = co_type_new (#x, x ## _new)), 1) \
	&& ((CO_DEBUG (1, "registered co type: " #x " = %d", COT (x))), 1)))

/// Declaration of action ID. Use in global library header file.
#define COP_DECL(x) extern unsigned __co_property_ ## x
/// Definition of action ID. Use in global part of .c file.
#define COP_DEF(x) unsigned __co_property_ ## x
/// Initialization of action ID. Use in static library init function.
#define COP_NEW(x) ((void)((((COP (x)) = co_property_new (#x)), 1) && \
	((CO_DEBUG (1, "registered co property: " #x " = %d", COP (x))), 1)))
/* }}} */ 
/* {{{ message handler helper macros */
/// Begin message handler switch.
#define CO_MH_BEGIN \
	CO_DEBUG (18, "entering message handler switch..."); \
	if (0)

/// Message handler switch case/option.
#define CO_MH_CASE(a, x) \
	else if (a == COM (x) \
			&& (CO_DEBUG (18, "chosen case: %s", #x), 1))

/// Message handler default switch case (MUST be last)
#define CO_MH_DEFAULT else

/* #define CO_MH_UNHANDLED(x) else (CO_DEBUG (1, \
			"WARNING: unhandled message %s", \
			__co_action_array [x])) */
#define CO_MH_UNHANDLED(x) else

/// End message handler switch (optional, IMHO).
#define CO_MH_END
/* }}} */

#define CO_OBJECT_IMPL(x, y)
#define CO_MSG_FW(x, y)
#define CO_MSG_FW_DEFAULT(x, y)

#ifdef __GNUC__
#define CO_POSSIBLY_UNUSED __attribute__ ((unused))
#else
#define CO_POSSIBLY_UNUSED
#endif

//#define CO_METH(x) x this, co self, co_id_t message
#define CO_METH(x, y) co_v x ## _ ## y \
	(x this CO_POSSIBLY_UNUSED, co self CO_POSSIBLY_UNUSED, \
	 co_id_t message CO_POSSIBLY_UNUSED, va_list ap CO_POSSIBLY_UNUSED)
#define CO_METH_INI(x, y,  ...) CO_METH_INI_ ## x ## _ ## y
#define CO_OBJECT(x) struct x; CO_TYPEDEF (x); struct x
//#define CO_ATTR(x, y) x y;
#define CO_ATTR(x, y) x y;

struct co;
typedef unsigned co_id_t;

typedef co (*co_constructor) ();

/// Function pointer to object message handler.
typedef co_v (*co_msg_h) (co o, co s, co_id_t a, va_list ap);
//typedef co_vr (*co_prop_h) (const co o, co_id_t p);

/* {{{ */
struct co_meth_entry
{
	co_id_t message;
	co_msg_h handler;
};
CO_TYPEDEF (co_meth_entry);
/* }}} */
/* {{{ */
struct co_meth_array
{
	size_t size;
	size_t n;
	co_meth_entry meth;
};
CO_TYPEDEF (co_meth_array);
/* }}} */

/* {{{ */
struct co_o_head
	/// Represents arbitrary object.
{
	/// Point to type_rec (which in turn may contain class reference)
	co_id_t type;
	/// Is set to 1 if object is being destroyed (internal)
	int destroy:1;
	/// Depth counter for message handler (internal)
	int msg_proc:17;
	/// Is set to 1 if list of listening objects should be sweeped.
	int check_listen:1;
	/// Is set to 1 if event is being processed.
	int ev_running:1;
};
CO_TYPEDEF (co_o_head);
/* }}} */
/* {{{ object head #define's */
#define CO_HEAD_MIN \
	co_o_head_s __co_root
#define CO_HEAD_SA \
	co_o_head_s __co_root; \
	co_msg_h __co_msg
#define CO_HEAD_FULL \
	co_o_head_s __co_root; \
	co_msg_h __co_msg; \
	co_meth_array __co_meth_ar; \
	struct co_ev_l *__co_listen
/* }}} */
/* {{{ */
struct co_o_head_full
{
	/// Point to type_rec (which in turn may contain class reference)
	co_id_t type;
	/// Is set to 1 if object is being destroyed (internal)
	int destroy:1;
	/// Is set to 1 if object is processing message right now (internal)
	int msg_proc:1;
	/// Is set to 1 if list of listening objects should be sweeped.
	int check_listen:1;
	/// Is set to 1 if event is being processed.
	int ev_running:1;
	co_msg_h __co_msg;
	co_meth_array __co_meth_ar;
	struct co_ev_l *__co_listen;
};
CO_TYPEDEF (co_o_head_full);
/* }}} */
/* {{{ */ struct co_ev_l
	/// Object connection record.
{
	/// Message ID to send if activated.
	co_id_t msg;
	/// Event ID, if raised connection is activated.
	co_id_t ev;
	/// Object to send message to.
	co o;
	/// Next list member.
	struct co_ev_l *n;
	int deleted:1;
	int ad:1;
};
CO_TYPEDEF (co_ev_l); /* }}} */

void *co_meth_array_new (void);
co_id_t co_meth_add (co_meth_array ar, co_id_t message, co_msg_h hendl);
co_v co_has_method (co o, co s, co_id_t a, va_list ap);

co_v co_m (co o, co s, co_id_t a, ...);
co_v co_m_v (co o, co s, co_id_t a, va_list ap);

void co_ev (co o, co_id_t ev, ...);

void co_o_destroy (co o);
void __co_del (co o);
void co_del (co o);
void co_del_f (co o);

void co_cnt (co emitor, co_id_t ev, co listener, co_id_t msg, bool ad);

void co_bless (co o, co_id_t t, co_msg_h msg);

void cox_co_cleanup (void *o, const void *);

co __co_o_bless (co o);
co __co_o_bless_p (co o);

#endif /* __CO_OBJECT_H__ */

#if 0
/* {{{ prop_get handler helper macros */
/// Begin message handler switch.
#define CO_PH_BEGIN if (0)
#define CO_PH_RET_C(p, x, t, v) \
	else if (p == COP (x)) do { \
		co_vr rv; \
		((const void *) (rv . t)) = &(v); \
		return rv; \
	} while (0)
#define CO_PH_RET(p, x, t, v) \
	else if (p == COP (x)) do { \
		co_vr rv; \
		rv . t = &(v); \
		return rv; \
	} while (0)
#define CO_PH_DEFAULT(x) else \
	{ \
		co_vr rv; \
		rv . _int = 0; \
		/*CO_DEBUG (1, "WARNING: unhandled property query %s",*/ \
		/*__co_property_array [x]);*/ \
		return rv; \
	}
/* }}} */
struct co_ln;
/* {{{ */ struct co_ln
	/// Element of co linked list.
{
	/// Associated object.
	co o;
	/// Next co_ln in list.
	struct co_ln *n;
	/// Previous co_ln in list.
	struct co_ln *p;
};
CO_TYPEDEF (co_ln); /* }}} */
/* {{{ */ struct co_o_type
{
	char *name;
	co_constructor cc;
}; /* }}} */
/* {{{ dynamic type id database */
extern struct co_o_type *__co_type_array;
extern unsigned __co_type_array_size;
extern char **__co_action_array;
extern unsigned __co_action_array_size;
extern char **__co_event_array;
extern unsigned __co_event_array_size;
extern char **__co_property_array;
extern unsigned __co_property_array_size;
/* }}} */
/* {{{ */ static inline co_constructor co_f_get (co_id_t t)
	/// Get object constructor from its type id.
	/**
	 * This function will return pointer to constructor for object of type @a
	 * t. You will barely use this function directly, macro CO_F_NEW is
	 * easier to use (and harder to misuse). If you call the constructor
	 * directly, don't forget to bless the object afterwards!
	 **/
{
	if (t < __co_type_array_size) {
		CO_DEBUG (3, "object factory: returning object %s for id %d",
				__co_type_array [t] . name, t);
		return __co_type_array [t] . cc;
	}
	return 0;
} /* }}} */
/* {{{ */ static inline co_constructor co_fs_get (const char *ts)
{
	co_id_t t;
	for (t = 0; t < __co_type_array_size; t ++) {
		if (! strcmp (__co_type_array [t] . name, ts)) {
			CO_DEBUG (3, "object factory: returning object %s for id %d",
					__co_type_array [t] . name, t);
			return __co_type_array [t] . cc;
		}
	}
	return 0;
} /* }}} */
co_id_t co_type_new (char *t, void *cc);
co_id_t co_action_new (char *a);
co_id_t co_event_new (char *);
co_id_t co_property_new (char *);

int _libco_static_init (void);
int _libco_static_free (void);
/* {{{ co actions */
/// Action sent to object after fully constructing it.
COM_DECL (bless);
/// Destroy object. NEVER send to object directly! Use co_del instead.
COM_DECL (destroy);
/// Disconnect object (all connections).
COM_DECL (disconnect_object);
/// Duplicate object, return duplicate.
COM_DECL (dup);
/// Draw object.
COM_DECL (draw);
/// Move object or its position pointer.
COM_DECL (move);
/// Update object (on display, not reload).
COM_DECL (update);
/// Hide object.
COM_DECL (hide);
/// Process incoming key (from keyboard).
COM_DECL (proc_key);
/// Add child object (independent of inheritance hierarchy).
COM_DECL (child_add);
/// Remove child object (independent of inheritance hierarchy).
COM_DECL (child_rem);
/// Scroll object (widget) to given position.
COM_DECL (scroll_to);
/// Scroll object (widget) by given offset(s).
COM_DECL (scroll_by);
/// Print object.
COM_DECL (print);
/// Save object (to disk, database, ...).
COM_DECL (save);
/// Load object state (from disk, database, ...).
COM_DECL (load);
/// Sort data in object
COM_DECL (sort);
/// Reload object data from file/db/etc...
COM_DECL (reload);
/// Execute action set up before.
COM_DECL (exec);
/// Query object (database)
COM_DECL (query);
/// Reset object to initial state.
COM_DECL (reset);
/// Append item to object.
COM_DECL (item_add);
/// Remove item from object.
COM_DECL (item_del);
/// Get object item.
COM_DECL (item_get);
/// Get type of sortvalue.
COM_DECL (sortt_get);
/// Get strval of item.
COM_DECL (item_str_get);
/// Run binary search on object.
COM_DECL (bsearch);
/// Find first occurance of key.
COM_DECL (ffind);
/// Get string value of object.
COM_DECL (strval);
/// Get sortvalue of object.
COM_DECL (sortval);
/// Create object in some external storage/etc.
COM_DECL (create);
/// Delete object from external storage/etc.
COM_DECL (delete);
COM_DECL (format);
COM_DECL (sp_get);
COM_DECL (reassign);
/* }}} */
/* {{{ co properties */
COP_DECL (nrows);
COP_DECL (ncols);
COP_DECL (nitems);
COP_DECL (header);
COP_DECL (name);
COP_DECL (width);
COP_DECL (col);
COP_DECL (type);
COP_DECL (sortval);
COP_DECL (id);
COP_DECL (flags);
COP_DECL (proto);
/* }}} */
/* {{{ co events */
/// Object was destroyed.
COE_DECL (destroyed);
// Fake event used by object destruction routines.
COE_DECL (disconnect_object);
/// OK event.
COE_DECL (ok);
/// Cancel event.
COE_DECL (cancel);
/// Insert event.
COE_DECL (insert);
/// Delete event.
COE_DECL (delete);
/// Delete event.
COE_DECL (edit);
/* }}} */
#endif /* 0 */
