/***{{{*******************************************************************
 * This file is part of libco - object library for C                     *
 * Copyright (c) 2002                                                    *
 *     Peter Rockai (yenar) <yenar@host.sk>                              *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2 of the License, or (at your option) any later version.      *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 59 Temple Place, Suite 330,                         *
 * Boston, MA 02111-1307  USA                                            *
 *******************************************************************}}}***/
/* {{{ file description */
/** \file co_type.h
 *  \brief Definition of multi-type variable types :). */
/* }}} */

#ifndef __CO_TYPE_H__
#define __CO_TYPE_H__

#include <stddef.h>


/// Define types for object, following libco conventions.
#define CO_TYPEDEF(x) typedef struct x x ## _s; typedef x ## _s *x
//#define CO_RETURN(x, y) do { co_v_t ret; ret . t = co_t_ ## x; ret . v . x = y; return ret; } while (0)
#define CO_RETURN(x, y) do { co_v ret; ret . _ ## x = y; return ret; } while (0)
#define CO_VAL(x, y) (x . _ ## y)

#define CO_T_DECL(x) co_t_ ## x, co_t_ ## x ## _p, co_t_ ## x ## _p_p
#define CO_V_DECL(x) x _ ## x; x ## _p _ ## x ## _p; x ## _p_p _ ## x ## _p_p
//#define CO_VTP_DECL(x) x *_ ## x ## _p;

/* {{{ */ enum co_type
/**
 * This enum lists C types that can be used in co_v or co_vr unions. You can
 * use them for other purposes too :). This should be fairly complete. The
 * names are mostly abbreviations from C type names, pointers are denoted by
 * _p suffix.
 **/
{
	CO_T_DECL (int),
	CO_T_DECL (long),
	CO_T_DECL (co),
	CO_T_DECL (void),
	CO_T_DECL (long_long),
	CO_T_DECL (double),
	CO_T_DECL (char),
	CO_T_DECL (co_type)
}; typedef enum co_type co_type; /* }}} */
/* {{{ simple typedefs */
typedef void *co;
#define _CO_TDEF(x) typedef x * x ## _p; typedef x ** x ## _p_p; typedef x *** x ## _p_p_p;
typedef long long long_long;
typedef unsigned long long unsigned_long_long;
typedef signed long long signed_long_long;
typedef unsigned long unsigned_long;
typedef signed long signed_long;
typedef long signed long_signed;
typedef long unsigned long_unsigned;
typedef unsigned short unsigned_short;
typedef signed short signed_short;
typedef short signed short_signed;
typedef short unsigned short_unsigned;

_CO_TDEF (void);
_CO_TDEF (int);
_CO_TDEF (char);
_CO_TDEF (long);
_CO_TDEF (size_t);
_CO_TDEF (co);
_CO_TDEF (long_long);
_CO_TDEF (double);
_CO_TDEF (co_type);
/* }}} */
/* {{{ */ union co_v
	/// Multi-type value.
	/**
	 * This is (on most machines?) 4-byte type, that can hold multitude of
	 * different same-sized or smaller types. Naming is same as in \ref
	 * co_type.  Larger types are available in form of pointers. This
	 * limitation is due to extensive usage of this type (eg. as return value
	 * of all co object message handlers).
	 **/
{
	CO_V_DECL (int);
	CO_V_DECL (long);
	CO_V_DECL (co);
	CO_V_DECL (void_p);
	CO_V_DECL (long_long_p);
	CO_V_DECL (double_p);
	CO_V_DECL (char);
	CO_V_DECL (co_type);
}; typedef union co_v co_v; /* }}} */

/* {{{ */ struct co_v_t
	/// Multi-type value with embedded type id.
	/**
	 * This is similar to co_v type, but it embeds type, so the type can be
	 * extracted at run time.
	 **/
{
	/// type of value stored in v
	co_type t;
	/// actual value
	co_v v;
};
typedef struct co_v_t co_v_t;
/* }}} */

/* {{{ */
struct co_type_rec {
	int id;
	int size;
	int ptr;
	const char *name;
};
/* }}} */

co_v co_v_copy (co_type, co_v);
char *co_v_format (co_type, co_v);

#if 0
/* {{{ pointer typedefs */
#define __CO_TPTR(x) typedef x *x ## _p
__CO_TPTR (void);
__CO_TPTR (void_p);
__CO_TPTR (long_long);
__CO_TPTR (unsigned_long);
__CO_TPTR (signed_long);
__CO_TPTR (co);
__CO_TPTR (double);
__CO_TPTR (char);
__CO_TPTR (char_p);
/* }}} */
/* {{{ */ union co_vr
	/// Multi-type reference to value.
	/**
	 * This is very similar to co_v, but instead of direct values pointers to
	 * them are used. Beware: _p types become pointers to pointers in this
	 * type!
	 **/
{
	CO_VR_DECL (int);
	CO_VR_DECL (long);
	CO_VR_DECL (co);
	CO_VR_DECL (void_p);
	CO_VR_DECL (long_long_p);
	CO_VR_DECL (double_p);
	CO_VR_DECL (char_p);
	CO_VR_DECL (char_p_p);
	CO_VR_DECL (co_type);
};
typedef union co_vr co_vr; /* }}} */
/* {{{ */ struct co_vr_t
	/// Multi-type value reference with embedded type id.
	/**
	 * This is similar to co_vr type, but it embeds type, so the type can be
	 * extracted at run time.
	 **/
{
	/// type of value referenced by v
	co_type t;
	/// actual value reference
	co_vr v;
};
typedef struct co_vr_t co_vr_t;
/* }}} */
#endif /* 0 */

#endif /* __CO_TYPE_H__ */
